import numpy as np
import time

from loguru import logger
import hgdl.misc as misc
import hgdl.local_methods as local
import hgdl.global_methods as glob
import dask.distributed as distributed
import dask.multiprocessing
from dask.distributed import as_completed

class optima:
    """
    stores all results and adaptations of it
    """
    def __init__(self,dim_x, max_optima):
        """
        input:
        -----
            dim ... the dimensionality of the space
            max_optima ... maximum number of stored optima
        """

        self.dim_x = dim_x
        self.max_optima = max_optima
        self.list = {"x": np.empty((0,self.dim_x)),
                     "f(x)": np.empty((0)),
                     "classifier": [],
                     "Hessian eigvals": np.empty((0,self.dim_x)),
                     "df/dx": np.empty((0,self.dim_x)),
                     "|df/dx|":np.empty((0)),
                     "success": False}
    ####################################################
    def fill_in_optima_list(self,res):
        x,f,fg,eig,local_success = res[0],res[1],res[2],res[3],res[4]
        if not np.any(local_success) and len(self.list["x"]) == 0: local_success[:] = True
        clean_indices = np.where(np.asarray(local_success) == True)[0]
        if len(clean_indices) == 0:
            return {"x": self.list["x"],
                    "f(x)": self.list["f(x)"],
                    "classifier": self.list["classifier"],
                    "Hessian eigvals": self.list["Hessian eigvals"],
                    "df/dx": self.list["df/dx"],
                    "|df/dx|": self.list["|df/dx|"],
                    "success": self.list["success"]}

        clean_x = x[clean_indices]
        clean_f = f[clean_indices]
        clean_fg = fg[clean_indices]
        clean_eig = eig[clean_indices]
        classifier = []
        for i in range(len(clean_x)):
            if any(clean_fg[i] > 1e-5): classifier.append("degenerate")
            elif any(np.isnan(clean_eig[i])): classifier.append("optimum")
            elif all(clean_eig[i] > 0.0): classifier.append("minimum")
            elif all(clean_eig[i] < 0.0): classifier.append("maximum")
            elif any(abs(clean_eig[i]) < 10e-6): classifier.append("zero curvature")
            elif len(np.where(clean_eig[i] < 0.0)[0])  < len(clean_eig[i]): classifier.append("saddle point")
            else: classifier.append("ERROR")

        optima_list =  {"x":       np.vstack([self.list["x"],clean_x]),
                        "f(x)":    np.append(self.list["f(x)"],clean_f),
                        "classifier":   self.list["classifier"] + classifier,
                        "Hessian eigvals": np.vstack([self.list["Hessian eigvals"],clean_eig]),
                        "df/dx":np.vstack([self.list["df/dx"],clean_fg]),
                        "|df/dx|":np.append(self.list["|df/dx|"],np.linalg.norm(clean_fg, axis = 1)),
                        "success": True}

        sort_indices = np.argsort(optima_list["f(x)"])
        optima_list["x"] = optima_list["x"][sort_indices][0:self.max_optima]
        optima_list["f(x)"] = optima_list["f(x)"][sort_indices][0:self.max_optima]
        optima_list["classifier"] = [optima_list["classifier"][i] for i in sort_indices][0:self.max_optima]
        optima_list["Hessian eigvals"] = optima_list["Hessian eigvals"][sort_indices][0:self.max_optima]
        optima_list["df/dx"] = optima_list["df/dx"][sort_indices][0:self.max_optima]
        optima_list["|df/dx|"] = optima_list["|df/dx|"][sort_indices][0:self.max_optima]
        self.list = dict(optima_list)
        return optima_list

    def get_minima(self,n):
        try:
            index = [i for i,x in enumerate(self.list["classifier"]) if x == "minimum"]
            index = index[0:min(n,len(index))]
            return self.list["x"][index], self.list["f(x)"][index]
        except:
            logger.debug("no minima available in the optima_list")
            return np.empty((0,self.dim_x)),np.empty((0))
    ####################################################
    def get_maxima(self,n):
        try:
            index = [i for i,x in enumerate(self.list["classifier"]) if x == "maximum"]
            index = index[0:min(n,len(index))]
            return self.list["x"][index], self.list["f(x)"][index]
        except:
            logger.debug("no maxima available in the optima_list")
            return np.empty((0,self.dim_x)),np.empty((0))
    ####################################################
    def get_deflation_points(self,n):
        try:
            index = [i for i, x in enumerate(self.list["classifier"]) if x == "maximum" or x == "minimum" or x == "zero curvature" or x == "saddle point" or x == "optimum"]
            return self.list["x"][index], self.list["f(x)"][index]
        except:
            logger.debug("no deflation points available in the optima_list")
            return np.empty((0,self.dim_x+self.dim_k)),np.empty((0))

#########################################################
#########################################################
#########################################################
#########################################################
#########################################################
class optima_constr:
    """
    stores all results and adaptations of it
    """
    def __init__(self,dim_x,dim_k, max_optima):
        """
        input:
        -----
            dim ... the dimensionality of the space
            max_optima ... maximum number of stored optima
        """

        self.dim_x = dim_x
        self.dim_k = dim_k
        self.dim = self.dim_x + self.dim_k

        self.max_optima = max_optima
        self.list = {"x": np.empty((0,self.dim_x)),
                     "k": np.empty((0,self.dim_k)),
                     "f(x)": np.empty((0)),
                     "classifier": [],
                     "Hessian eigvals": np.empty((0,self.dim_x)),
                     "dL/dx": np.empty((0,self.dim_x)),
                     "dL/dk": np.empty((self.dim_k)),
                     "|dL/dx|":np.empty((0)),
                     "|dL/dk|":np.empty((0)),
                     "success": False}
    ####################################################
    def fill_in_optima_list(self,res):
        x,f,Lg,eig,local_success = res[0],res[1],res[2],res[3],res[4]
        eig = eig[:,0:self.dim_x]
        if not np.any(local_success) and len(self.list["x"]) == 0: local_success[:] = True
        clean_indices = np.where(np.asarray(local_success) == True)[0]

        if len(clean_indices) == 0:
            return {"x": self.list["x"][:,0:self.dim_x],
                    "k": self.list["k"][:,self.dim_x:self.dim],
                    "f(x)": self.list["f(x)"],
                    "classifier": self.list["classifier"],
                    "Hessian eigvals": self.list["Hessian eigvals"],
                    "dL/dx": self.list["dL/dx"],
                    "dL/dk": self.list["dL/dk"],
                    "|dL/dx|": self.list["|dL/dx|"],
                    "|dL/dk|": self.list["|dL/dk|"],
                    "success": self.list["success"]}

        clean_x = x[clean_indices]
        clean_f = f[clean_indices]
        clean_Lg = Lg[clean_indices]
        clean_eig = eig[clean_indices]
        classifier = []
        for i in range(len(clean_x)):
            if any(clean_Lg[i] > 1e-5): classifier.append("degenerate")
            elif any(np.isnan(clean_eig[i])): classifier.append("optimum")
            elif all(clean_eig[i] > 0.0): classifier.append("minimum")
            elif all(clean_eig[i] < 0.0): classifier.append("maximum")
            elif any(abs(clean_eig[i]) < 10e-6): classifier.append("zero curvature")
            elif len(np.where(clean_eig[i] < 0.0)[0])  < len(clean_eig[i]): classifier.append("saddle point")
            else: classifier.append("ERROR")


        optima_list =  {"x":       np.vstack([self.list["x"],clean_x[:,0:self.dim_x]]),
                        "k":       np.vstack([self.list["k"],clean_x[:,self.dim_x:self.dim]]),
                        "f(x)":    np.append(self.list["f(x)"],clean_f),
                        "classifier":   self.list["classifier"] + classifier,
                        "Hessian eigvals": np.vstack([self.list["Hessian eigvals"],clean_eig]),
                        "dL/dx":np.vstack([self.list["dL/dx"],clean_Lg[:,0:self.dim_x]]),
                        "|dL/dx|":np.append(self.list["|dL/dx|"],np.linalg.norm(clean_Lg[:,0:self.dim_x],axis = 1)),
                        "dL/dk":np.vstack([self.list["dL/dk"],clean_Lg[:,self.dim_x:self.dim]]),
                        "|dL/dk|":np.append(self.list["|dL/dk|"],np.linalg.norm(clean_Lg[:,self.dim_x:self.dim],axis = 1)),
                        "success": True}

        sort_indices = np.argsort(optima_list["f(x)"])
        optima_list["x"] = optima_list["x"][sort_indices][0:self.max_optima]
        optima_list["k"] = optima_list["k"][sort_indices][0:self.max_optima]
        optima_list["f(x)"] = optima_list["f(x)"][sort_indices][0:self.max_optima]
        optima_list["classifier"] = [optima_list["classifier"][i] for i in sort_indices][0:self.max_optima]
        optima_list["Hessian eigvals"] = optima_list["Hessian eigvals"][sort_indices][0:self.max_optima]
        optima_list["dL/dx"] = optima_list["dL/dx"][sort_indices][0:self.max_optima]
        optima_list["dL/dk"] = optima_list["dL/dk"][sort_indices][0:self.max_optima]
        optima_list["|dL/dx|"] = optima_list["|dL/dx|"][sort_indices][0:self.max_optima]
        optima_list["|dL/dk|"] = optima_list["|dL/dk|"][sort_indices][0:self.max_optima]
        self.list = dict(optima_list)
        return optima_list
    ####################################################
    def get_minima(self,n):
        try:
            index = [i for i,x in enumerate(self.list["classifier"]) if x == "minimum"]
            index = index[0:min(n,len(index))]
            return self.list["x"][index], self.list["f(x)"][index]
        except:
            logger.debug("no minima available in the optima_list")
            return np.empty((0,self.dim_x)),np.empty((0))
    ####################################################
    def get_maxima(self,n):
        try:
            index = [i for i,x in enumerate(self.list["classifier"]) if x == "maximum"]
            index = index[0:min(n,len(index))]
            return self.list["x"][index], self.list["f(x)"][index]
        except:
            logger.debug("no maxima available in the optima_list")
            return np.empty((0,self.dim_x)),np.empty((0))
    ####################################################
    def get_deflation_points(self,n):
        try:
            index = [i for i, x in enumerate(self.list["classifier"]) if x == "maximum" or x == "minimum" or x == "zero curvature" or x == "saddle point" or x == "optimum"]
            return np.column_stack([self.list["x"][index],self.list["k"][index]]), self.list["f(x)"][index]
        except:
            logger.debug("no deflation points available in the optima_list")
            return np.empty((0,self.dim_x+self.dim_k)),np.empty((0))




import numpy as np
import hgdl.misc as misc
import hgdl.local_methods.bump_function as defl
import dask.distributed as distributed
from loguru import logger


def DNewton(func,grad,hess,bounds,x0,max_iter,tol,*args):
    e = np.inf
    gradient = np.ones((len(x0))) * np.inf
    counter = 0
    x = np.array(x0)
    grad_list = []
    while e > tol or np.max(abs(gradient)) > tol:
        x = misc.project_onto_bounds(x,bounds)
        x[abs(x)<1e-16] = 0.
        gradient = grad(x,*args)
        gradient[abs(gradient)<1e-16] = 0.
        if hess: hessian  = hess(x,*args)
        else: hessian = approximate_hessian(x,grad,*args)
        hessian[abs(hessian) < 1e-16] = 0.
        grad_list.append(np.max(gradient))
        try: gamma = np.linalg.solve(hessian,-gradient)
        except Exception as error: gamma,a,b,c = np.linalg.lstsq(hessian,-gradient,rcond=None)
        if any(gamma == np.nan) or any(gamma == np.inf): return x,func(x, *args),gradient,np.linalg.eig(hess(x, *args))[0], False
        x += gamma
        e = np.max(abs(gamma))
        logger.debug("dNewton step size: ", e, " max gradient: ",np.max(abs(gradient)))
        #print("dNewton step size: ", e, " max gradient: ",np.max(abs(gradient)))
        if counter > max_iter: return x,func(x, *args),gradient,np.linalg.eig(hess(x, *args))[0], False, "max_iter reached"
        counter += 1
    return x,func(x, *args),gradient,np.linalg.eig(hess(x, *args))[0], True, "converged"



def approximate_hessian(x, grad, *args):
    ##implements a first-order approximation
    len_x = len(x)
    hess = np.zeros((len_x,len_x))
    epsilon = 1e-6
    grad_x = grad(x, *args)
    for i in range(len_x):
        x_temp = np.array(x)
        x_temp[i] = x_temp[i] + epsilon
        hess[i,i:] = ((grad(x_temp,*args) - grad_x)/epsilon)[i:]
    return hess + hess.T - np.diag(np.diag(hess))










import numpy as np


class meta_data:
    """
    this class contains all the data necessary to run the workers
    """
    def __init__(self,obj):
        self.workers = obj.workers    ###dictionary of host and walker workers
        self.x0 = obj.x0
        self.func = obj.func
        self.grad = obj.grad
        self.hess = obj.hess
        self.L = obj.L
        self.Lgrad = obj.Lgrad
        self.Lhess = obj.Lhess
        self.bounds = obj.bounds
        self.radius = obj.radius
        self.dim_x = obj.dim_x
        self.dim_k = obj.dim_k
        self.dim = obj.dim_x + self.dim_k

        self.local_max_iter = obj.local_max_iter
        self.number_of_walkers = obj.number_of_walkers
        self.num_epochs = obj.num_epochs
        self.global_optimizer = obj.global_optimizer
        self.local_optimizer = obj.local_optimizer
        self.args = obj.args
        self.constr = obj.constr
        self.tolerance = obj.tolerance



import numpy as np
import time

def out_of_bounds(x,bounds):
    for i in range(len(x)):
        if x[i] < bounds[i,0] or x[i] > bounds[i,1]:
            return True
    return False

def in_bounds(x, bounds):
    if (bounds[:,1]-x > 0).all() and (bounds[:,0] - x < 0).all():
        return True
    return False

def random_sample(N,k,bounds):
    sample = np.random.random((N, k))
    sample *= bounds[:,1] - bounds[:,0]
    sample += bounds[:,0]
    return sample

def random_population(bounds, n):
    return np.random.uniform(low = bounds[:,0], high = bounds[:,1], size = (n,len(bounds)))

def finish_up_tasks(tasks):
    for f in tasks:
        if f.status == 'cancelled':
            tasks.remove(f)
    while any(f.status == 'pending' for f in tasks):
        time.sleep(0.1)
    return tasks

def project_onto_bounds(x,bounds):
    new_x = np.array(x)
    for i in range(len(x)):
        if x[i] < bounds[i,0]: new_x[i] = bounds[i,0]
        if x[i] > bounds[i,1]: new_x[i] = bounds[i,1]
    return new_x





import numpy as np
import time

from loguru import logger

import hgdl.misc as misc
import dask.distributed
from distributed import Client, get_client, secede, rejoin, protocol
import dask.distributed as distributed
from hgdl.local_methods.dNewton import DNewton
from scipy.optimize import minimize
import hgdl.local_methods.bump_function as defl


def run_local(d,optima,x0):
    x_defl,f_defl = optima.get_deflation_points(len(optima.list))
    return run_local_optimizer(d,x0,x_defl)
###########################################################################
def run_local_optimizer(d,x0,x_defl = []):
    """
    this function runs a deflated local methos for
    all the walkers.
    The loop below goes over every walker
    input:
        2d numpy array of initial positions
        2d numpy array of positions of deflations (optional, default = [])
    return:
        optima_locations, func values, gradient norms, eigenvalues, local_success(bool)
    """
    dim = d.dim
    number_of_walkers = d.number_of_walkers

    if len(x0) < number_of_walkers:
        x0 = np.row_stack([x0,misc.random_population(d.bounds,number_of_walkers - len(x0))])

    client = get_client()
    tasks = []
    bf = client.scatter(d,workers = d.workers["walkers"])
    for i in range(min(len(x0),number_of_walkers)):
        logger.debug("Worker ",i," submitted")
        #print("Worker ",i," submitted", flush = True)
        worker = d.workers["walkers"][(int(i - ((i // number_of_walkers) * number_of_walkers)))]
        data = {"d":bf,"x0":x0[i],"x_defl":x_defl}
        tasks.append(client.submit(local_method,data,workers = worker))

    results = client.gather(tasks)
    number_of_walkers = len(tasks)
    x = np.empty((number_of_walkers, dim))
    f = np.empty((number_of_walkers))
    Lg = np.empty((number_of_walkers, dim))
    eig = np.empty((number_of_walkers,dim))
    local_success = np.empty((number_of_walkers), dtype = bool)

    for i in range(len(tasks)):
        x[i],f[i],Lg[i],eig[i],local_success[i] = results[i]
        client.cancel(tasks[i])
        for j in range(i):
            if np.linalg.norm(np.subtract(x[i],x[j])) < 2.0 * d.radius and local_success[j] == True:
                logger.warning("points converged too close to each other in HGDL; point removed")
                local_success[j] = False; break
        for j in range(len(x_defl)):
            if np.linalg.norm(np.subtract(x[i],x_defl[j])) < 2.0 * d.radius and all(Lg[i] < 1e-5):
                logger.warning("local method converged within 2 x radius of a deflated position in HGDL")
                local_success[i] = False
    return x, f, Lg, eig, local_success

def local_method(data, method = "dNewton"):
    from functools import partial
    d = data["d"]
    x0 = np.array(data["x0"])
    e = np.inf
    local_success = True
    tol = d.tolerance
    x_defl = data["x_defl"]
    bounds = d.bounds
    max_iter = d.local_max_iter
    args = d.args
    method = d.local_optimizer
    #augment grad, hess
    Lgrad = partial(defl.deflated_grad, grad_func = d.Lgrad, x_defl = x_defl, radius = d.radius)
    if callable(d.Lhess):
        Lhess = partial(defl.deflated_hess, grad_func = d.Lgrad,
                       hess_func = d.Lhess, x_defl = x_defl, radius = d.radius)
    else:
        Lhess = d.Lhess

    #call local methods
    if method == "dNewton":
        x,L,Lg,eig,local_success = DNewton(d.L,Lgrad,Lhess,bounds,x0,max_iter,tol,*args)
        f = d.func(x,*args)

    elif type(method) == str:
        if d.constr: logger.warning("Deflated constraints are only supported by local optimizer 'DNewton'")
        res = minimize(d.func, x0, args = args, method = method, jac = Lgrad, hess = Lhess,
                       bounds = bounds, options = {"disp":False})
        x = res["x"]
        f = res["fun"]
        eig = np.ones(x.shape) * np.nan
        Lg = res["jac"]
        local_success = res["success"]

    elif callable(method):
        res = method(d.func,Lgrad,Lhess,bounds,x0,*args)
        x = res["x"]
        f = res["fun"]
        eig = np.ones(x.shape) * np.nan
        Lg = res["jac"]
        local_success = res["success"]


    else: raise Exception("no local method specified")

    return x,f,Lg,np.real(eig),local_success
###########################################################################



import numpy as np

class NonLinearConstraint:
    """
    This class provides HGDL with the capability of constrained function
    optimization. Every constraint should be fined by one class instance.

    Parameters:
    -----------
    nlc : Callable
        The function of the constraint.
    nlc_grad : Callable
        The gradient of the constraint.
    nlc_hess : Callable, optional
        The Hessian of the constraint, if available.
        Default = None. In this case it will be approximated.
        If no Hessian is available, don't give HGDL the Hessian of the objective function.
    ctype : str, optional
        String '<', '>', or '=', defining the constraint as inequality or equalit constraint.
        Default = '='.
    values : float, optional
        Value of the constraint. Default = 0.0.
    """

    def __init__(self,nlc,nlc_grad, nlc_hess = None, ctype = "=", value = 0.0, bounds = np.array([[-1e6,1e6]])):
        self.nlc = nlc
        self.nlc_grad = nlc_grad
        self.nlc_hess = nlc_hess
        self.ctype = ctype
        self.value = value
        self.bounds = bounds
        self.multiplier_index = None
        self.slack_index = None
        if ctype != "=" and len(self.bounds) == 1: self.bounds = np.array([[-1e6,1e6],[-1e6,1e6]])

    def set_multiplier_index(self,index):
        self.multiplier_index = index
    def set_slack_index(self,index):
        self.slack_index = index




import numpy as np
import time

from loguru import logger
from functools import *
import hgdl.misc as misc
from hgdl.local_methods.local_optimizer import run_local_optimizer
from hgdl.global_methods.global_optimizer import run_global
from hgdl.local_methods.local_optimizer import run_local
import dask.distributed as distributed
import dask.multiprocessing
from dask.distributed import as_completed
from hgdl.optima  import optima
from hgdl.optima  import optima_constr
from hgdl.meta_data  import meta_data
import pickle


class HGDL:
    """
    This is HGDL, a class for asynchronous HPC-capable optimization. \n
    H ... Hybrid \n
    G ... Global \n
    D ... Deflated \n
    L ... Local \n
    The algorithm places a number of walkers inside the domain (the number is determined by the dask client), all of which perform
    a local optimization in a distributed way in parallel. When the walkers have identified local optima, their positions are communicated back to the host
    who removes the found optima by deflation, and replaces the fittest walkers by a global optimization step. From here the next epoch
    begins with distributed local optimizations of the new walkers. The algorithm results in a sorted list of unique optima (only if optima
    are of the form f'(x) = 0)
    The method `hgdl.optimize` instantly returns a result object that can be queried for a growing, sorted list of
    optima. If a Hessian is provided, those optima are classified as minima, maxima or saddle points.

    Parameters
    ----------
    func : Callable
        The function to be MINIMIZED. A callable that accepts an np.ndarray and optional arguments, and returns a scalar.
    grad : Callable
        The gradient of the function to be MINIMIZED. A callable that accepts an np.ndarray and optional arguments, and returns a vector
        (np.ndarray) of shape (D), where D is the dimensionality of the space in which the
        optimization takes place. Remeber that D changes if constraints are used; account for the derivatives with respect to multipliers and
        slack variables.
    bounds : np.ndarray
        The bounds of the domain; an np.ndarray of shape (D x 2), where D is the dimensionality of the space in which the
        optimization takes place. Here D is the dimension of the original input space without slack variables or multipliers. Those bounds
        can be specified when seeting upt he constraint.
    hess : Callable, optional
        The Hessian of the function to be MINIMIZED. A callable that accepts an np.ndarray and optional arguments, and returns a
        np.ndarray of shape (D x D). The default value is no-op. If provided the Hessian should also include the derivatives with respect to
        slack variables and multipliers, if constraints are used.
    num_epochs : int, optional
        The number of epochs the algorithm runs through before being terminated. One epoch is the convergence of all local walkers,
        the deflation of the identified optima, and the global replacement of the walkers. Note, the algorithm is running asynchronously, so a high number
        of epochs can be chosen without concern, it will not affect the run time to obtain the optima. Therefore, the default is
        100000.
    global_optimizer : Callable or str, optional
        The function (identified by a string or a Callable) that replaces the fittest walkers after their local convergence.
        The possible options are `genetic` (default), `gauss`, `random` or a callable that accepts an
        np.ndarray of shape (U x D) of positions, an np.ndarray of shape (U) of function values,
        and np.ndarray of shape (D x 2) of bounds, and an integer specifying the number of offspring
        individuals that should be returned. The callable should return the positions of the offspring
        individuals as an np.ndarray of shape (number_of_offspring x D).
    local_optimizer : Callable or str, optional
        The local optimizer that is used. The options are
        `dNewton` (default), `L-BFGS-B`, `BFGS`, `CG`, `Newton-CG`.
        The above methods have been tested, but most others should work. Visit the `scipy.optimize.minimize` docs for specifications
        and limitations of the local methods. The parameter also accepts a callable that accepts as input a function, gradient, Hessian,
        bounds (all as specified above), and args, and returns an object similar to the scipy.optimize.minimize methods. `DNewton` should be used
        for constrained optimization.
    number_of_optima : int, optional
        The number of optima that will be stored in the optima list and deflated. The default is 1e6.
        After that number is reached, worse-performing optima will not be stored or deflated.
    radius: float, optional
        The radius of the deflation operator. The default is estimated from the size of the domain.
        This will be changed in future releases to be estimated from the curvature of the function
        at the optima.
    local_max_iter : int, optional
        The number of iterations before local optimizations are terminated. The default is 1000.
        It can be lowered when second-order local optimizers are used.
    args : tuple, optional
        A tuple of arguments that will be communicated to the function, the gradient, and the Hessian callables.
        Default = ().
    constr : object, optional
        An optional n-tuple of constraint objects. See hgdl.constraints.NonLinearConstraints.
        The default is no constraints (). Make sure you use a `local_optimizer = "dNewton"` if constraints are used.
        Also, try to provide a Hessian callable. If not, the Hessian will be approximated from the gradient which is fine as long as
        your gradient function is fast and exact. Don't bother with a constraint Hessian if you don't have the objective function Hessian; it won't be used.
        When constraints are used, the space is extended by a linear space with elements 'k' which
        are the Lagrangian multipliers and slack variables and will be part of the result object.

    Attributes
    ----------
    optima : object
        Contains the attribute optima.list in which the optima are stored. 
        However, the method 'get_latest()' should be used to access the optima.


    """

    def __init__(self, func, grad, bounds,
            hess = None, num_epochs=100000,
            global_optimizer = "genetic",
            local_optimizer = "dNewton",
            number_of_optima = 1000000,
            radius = None,
            local_max_iter = 1000,
            args = (), constraints = ()):

        self.constr = constraints
        self.dim_x = len(bounds)
        self.func = func
        self.grad = grad
        self.hess = hess
        self.L = self.lagrangian
        self.Lgrad = self.lagrangian_grad
        self.Lhess = self.lagrangian_hess
        index = self.dim_x
        if self.constr and local_optimizer != "dNewton": raise Exception("Please use ``local_optimizer = 'dNewton' '' if constraints are used.")
        if local_optimizer == "dNewton": print("Warning: dNewton will not adhere to bounds. It is recommended to formulate your objective function such that it is defined on R^N by simple non-linear transformations.")
        for c in self.constr:
            if c.ctype == "=":
                c.set_multiplier_index(index)
                index += 1
            else:
                c.set_multiplier_index(index)
                c.set_slack_index(index+1)
                index += 2
            bounds = np.row_stack([bounds,c.bounds])

        self.bounds = np.asarray(bounds)
        if radius is None: self.radius = np.min(bounds[0:self.dim_x,1]-bounds[0:self.dim_x,0])/10000.0
        else: self.radius = radius
        self.dim = len(self.bounds)
        self.dim_k = self.dim - self.dim_x
        self.local_max_iter = local_max_iter
        self.num_epochs = num_epochs
        self.global_optimizer = global_optimizer
        self.local_optimizer = local_optimizer
        self.args = args
        if self.dim_k == 0: self.optima = optima(self.dim_x,number_of_optima)
        if self.dim_k >  0: self.optima = optima_constr(self.dim_x, self.dim_k,number_of_optima)
        logger.debug("HGDL successfully initiated {}")
        logger.debug("deflation radius set to {}", self.radius)
        if callable(self.hess): logger.debug("Hessian was provided by the user: {}", self.hess)
        logger.debug("========================")
    ###########################################################################
    ###########################################################################
    ############USER FUNCTIONS#################################################
    ###########################################################################
    ###########################################################################
    ###########################################################################
    def optimize(self, dask_client = None, x0 = None, tolerance = 1e-6):
        """
        Function to start the optimization. Note, this function will not return anything.
        Use the method hgdl.HGDL.get_latest() (non-blocking) or hgdl.HGDL.get_final() (blocking)
        to query results.

        Parameters
        ----------
        dask_client : distributed.client.Client, optional
            The client that will be used for the distibuted local optimizations. The default is a local client.
        x0 : np.ndarray, optional
            An np.ndarray of shape (V x D) of points used as starting positions.
            If V > number of walkers (specified by the dask client) the array will be truncated.
            If V < number of walkers, random points will be appended.
            The default is None, meaning only random points will be used.
        tolerance : float, optional
            The tolerance used by the local optimizers. The default is 1e-6
        """

        client = self._init_dask_client(dask_client)
        self.tolerance = tolerance
        logger.debug(client)
        self.x0 = self._prepare_starting_positions(x0)
        logger.debug("HGDL starts with: {}", self.x0)
        self.meta_data = meta_data(self)
        self._run_epochs(client)

    ###########################################################################
    def get_client_info(self):
        """
        Function to receive info about the workers.
        """
        return self.workers
    ###########################################################################
    def get_latest(self):
        """
        Function to request the current result.
        No inputs
        """
        try:
            data, frames = self.transfer_data.get()
            self.optima = distributed.protocol.deserialize(data,frames)
            logger.debug("HGDL called get_latest() successfully")
        except Exception as err:
            self.optima = self.optima
            logger.error("HGDL get_latest failed due to {} \n optima list unchanged", str(err))
        optima_list = self.optima.list
        return optima_list
    ###########################################################################
    def get_final(self):
        """
        Function to request the final result.
        CAUTION: This function will block the main thread until
        the end of all epochs is reached.
        No inputs.
        """
        try:
            self.optima = self.main_future.result()
        except Exception as err:
            logger.error("HGDL get_final failed due to {}", str(err))
        optima_list = self.optima.list
        return optima_list
    ###########################################################################
    def cancel_tasks(self):
        """
        Function to cancel all tasks and therefore the execution.
        However, this function does not kill the client.
        """
        logger.debug("HGDL is cancelling all tasks...")
        res = self.get_latest()
        self.break_condition.set(True)
        self.client.cancel(self.main_future)
        logger.debug("Status of HGDL task: ", self.main_future.status)
        logger.debug("This leaves the client alive.")
        return res
    ###########################################################################
    def kill_client(self):
        """
        Function to cancel all tasks and kill the dask client, and therefore the execution.
        If cancel_tasks() is called before, this will throw an error.
        """
        logger.debug("HGDL kill client initialized ...")
        res = self.get_latest()
        try:
            self.break_condition.set(True)
            self.client.cancel(self.main_future)
            self.client.close()
            logger.debug("HGDL kill client successful")
        except Exception as err:
            raise RuntimeError("HGDL kill failed") from err
        return res
    ###########################################################################
    ############USER FUNCTIONS END#############################################
    ###########################################################################
    def _prepare_starting_positions(self,x0):
        if x0 is not None and len(x0[0]) != self.dim: x0 = misc.random_population(self.bounds,self.number_of_walkers)
        if x0 is None: x0 = misc.random_population(self.bounds,self.number_of_walkers)
        elif x0.ndim == 1: x0 = np.array([x0])

        if len(x0) < self.number_of_walkers:
            x0_aux = np.zeros((self.number_of_walkers,len(x0[0])))
            x0_aux[0:len(x0)] = x0
            x0_aux[len(x0):] = misc.random_population(self.bounds,self.number_of_walkers - len(x0))
            x0 = x0_aux
        elif len(x0) > self.number_of_walkers:
            x0 = x0[0:self.number_of_walkers]
        else: x0 = x0
        return x0
    ###########################################################################
    def _init_dask_client(self,dask_client):
        if dask_client is None: 
            dask_client = dask.distributed.Client()
            logger.debug("No dask client provided to HGDL. Using the local client")
        else:
            logger.debug("dask client provided to HGDL")
        client = dask_client
        worker_info = list(client.scheduler_info()["workers"].keys())
        if not worker_info: raise Exception("No workers available")
        self.workers = {"host": worker_info[0],
                "walkers": worker_info[1:]}
        logger.debug(f"Host {self.workers['host']} has {len(self.workers['walkers'])} workers.")
        self.number_of_walkers = len(self.workers["walkers"])
        return client
    ###########################################################################
    def _run_epochs(self,client):
        self.break_condition = distributed.Variable("break_condition",client)
        self.transfer_data = distributed.Variable("transfer_data",client)
        a = distributed.protocol.serialize(self.optima)
        self.transfer_data.set(a)
        self.break_condition.set(False)
        data = {"transfer data":self.transfer_data,
                "break condition":self.break_condition,
                "optima":self.optima, "metadata":self.meta_data}
        bf = client.scatter(data, workers = self.workers["host"])
        self.main_future = client.submit(hgdl, bf, workers = self.workers["host"])
        self.client = client
    ###########################################################################
    def lagrangian(self,x, *args):
        L = self.func(x[0:self.dim_x], *args)
        for c in self.constr:
            if   c.ctype == '=': L += x[c.multiplier_index] * (c.nlc(x[0:self.dim_x],*args) - c.value)
            elif c.ctype == '<': L += x[c.multiplier_index] * (c.nlc(x[0:self.dim_x],*args) - c.value + x[c.slack_index]**2)
            elif c.ctype == '>': L += x[c.multiplier_index] * (c.nlc(x[0:self.dim_x],*args) - c.value - x[c.slack_index]**2)
            else: raise Exception("Wrong ctype in constraint")
        return L

    def lagrangian_grad(self,x, *args):
        Lgrad = np.zeros((len(x)))
        Lgrad[0:self.dim_x] = self.grad(x[0:self.dim_x], *args)
        for c in self.constr:
            Lgrad[0:self.dim_x] += x[c.multiplier_index] * c.nlc_grad(x[0:self.dim_x], *args)
            if c.ctype == '=':
                Lgrad[c.multiplier_index] = c.nlc(x[0:self.dim_x],*args) - c.value
            elif c.ctype == '<':
                Lgrad[c.multiplier_index] = c.nlc(x[0:self.dim_x],*args) - c.value + x[c.slack_index]**2
                Lgrad[c.slack_index] = 2.0 * x[c.multiplier_index] * x[c.slack_index]
            elif c.ctype == '>':
                Lgrad[c.multiplier_index] = c.nlc(x[0:self.dim_x],*args) - c.value - x[c.slack_index]**2
                Lgrad[c.slack_index] = -2.0 * x[c.multiplier_index] * x[c.slack_index]
            else: raise Exception("Wrong ctype in constraint")
        return Lgrad

    def lagrangian_hess(self,x, *args):
        Lhess = np.zeros((len(x),len(x)))
        if not self.hess: return self.lagrangian_hess_approx(x,*args)
        Lhess[0:self.dim_x,0:self.dim_x] = self.hess(x[0:self.dim_x], *args)
        for c in self.constr:
            Lhess[0:self.dim_x,0:self.dim_x] += x[c.multiplier_index] * c.nlc_hess(x[0:self.dim_x], *args)
            if c.ctype == '=':
                #mixed: x, multiplier
                cc = c.nlc_grad(x[0:self.dim_x],*args)
                Lhess[c.multiplier_index,0:self.dim_x] = cc
                Lhess[0:self.dim_x,c.multiplier_index] = cc
                #same
                ##Lhess[c.multiplier_index,c.multiplier_index] = 0 ##already zero
            elif c.ctype == '<':
                #mixed: x,multiplier
                cc = c.nlc_grad(x[0:self.dim_x],*args)
                Lhess[c.multiplier_index,0:self.dim_x] = cc
                Lhess[0:self.dim_x,c.multiplier_index] = cc
                #mixed: multiplier, slack
                cc = 2.0 * x[c.slack_index]
                Lhess[c.slack_index,c.multiplier_index] = cc
                Lhess[c.multiplier_index,c.slack_index] = cc
                #same: slack, slack
                cc = 2.0 * x[c.multiplier_index]
                Lhess[c.slack_index,c.slack_index] = cc

            elif c.ctype == '>':
                #mixed: x,multiplier
                cc = c.nlc_grad(x[0:self.dim_x],*args)
                Lhess[c.multiplier_index,0:self.dim_x] = cc
                Lhess[0:self.dim_x,c.multiplier_index] = cc
                #mixed: multiplier, slack
                cc = -2.0 * x[c.slack_index]
                Lhess[c.slack_index,c.multiplier_index] = cc
                Lhess[c.multiplier_index,c.slack_index] = cc
                #same: slack, slack
                cc = -2.0 * x[c.multiplier_index]
                Lhess[c.slack_index,c.slack_index] = cc

            else: raise Exception("Wrong ctype in constraint")
        return Lhess

    def lagrangian_hess_approx(self,x, *args):
        ##implements a first-order approximation
        #grad = self.lagrangian_grad
        len_x = len(x)
        hess = np.zeros((len_x,len_x))
        epsilon = 1e-6
        grad_x = self.lagrangian_grad(x, *args)
        for i in range(len_x):
            x_temp = np.array(x)
            x_temp[i] = x_temp[i] + epsilon
            hess[i,i:] = ((self.lagrangian_grad(x_temp,*args) - grad_x)/epsilon)[i:]
        return hess + hess.T - np.diag(np.diag(hess))
###########################################################################
###########################################################################
##################hgdl functions###########################################
###########################################################################
###########################################################################
def hgdl(data):
    metadata = data["metadata"]
    transfer_data = data["transfer data"]
    break_condition = data["break condition"]
    optima = data["optima"]
    logger.debug("HGDL computing epoch 1 of {}", metadata.num_epochs)
    res = run_local(metadata,optima,metadata.x0)
    optima.fill_in_optima_list(res)
    a = distributed.protocol.serialize(optima)
    transfer_data.set(a)

    for i in range(1,metadata.num_epochs):
        bc = break_condition.get()
        if bc is True:
            logger.debug(f"HGDL Epoch {i} was cancelled")
            break
        logger.debug(f"HGDL computing epoch {i+1} of ", metadata.num_epochs)
        optima = run_hgdl_epoch(metadata,optima)
        a = distributed.protocol.serialize(optima)
        transfer_data.set(a)
    logger.debug("HGDL finished all epochs!")
    return optima
###########################################################################
def run_hgdl_epoch(metadata,optima):
    optima_list = optima.list
    n = min(len(optima_list["x"]),metadata.number_of_walkers)
    global_res = run_global(\
            np.array(optima_list["x"][0:n]),
            np.array(optima_list["f(x)"][0:n]),
            metadata.bounds[0:metadata.dim_x], metadata.global_optimizer,n)
    x0 = np.zeros((n,metadata.dim))
    x0[:,0:metadata.dim_x] = np.array(global_res)
    if metadata.dim != metadata.dim_x: x0[:,metadata.dim_x:] = np.array(optima_list["k"][0:n])
    res = run_local(metadata,optima,x0)
    op = optima.fill_in_optima_list(res)
    return optima
